@@ -1,6 +1,6 @@
 /*
  * heap_malloc.c - use malloc heap as a main heap. (for test use)
- * @copyright (c) 2010, Tohoku University.
+ * @copyright (C) 2021 SML# Development Team.
  * @author UENO Katsuhiro
  */
 
@@ -53,7 +53,7 @@ sml_heap_gc()
 	sml_malloc_sweep(MAJOR);
 	RUN_THE_WORLD();
 	GIANT_UNLOCK();
-	sml_run_finalizer(NULL);
+	sml_run_finalizer();
 }
 
 SML_PRIMITIVE void *
