@@ -1,6 +1,6 @@
 /*
  * heap_otomo.c
- * @copyright (c) 2010, Tohoku University.
+ * @copyright (C) 2021 SML# Development Team.
  * @author OTOMO Toshiaki
  * @version $Id: $
  */
@@ -1102,7 +1102,7 @@ sml_heap_gc(void)
   
 #ifndef FAIR_COMPARISON
   /* start finalizers */
-  sml_run_finalizer(NULL);
+  sml_run_finalizer();
 #endif /* FAIR_COMPARISON */
 }
 
