@@ -1,10 +1,13 @@
 /**
  * prim.c
- * @copyright (c) 2007-2010, Tohoku University.
+ * @copyright (C) 2021 SML# Development Team.
  * @author UENO Katsuhiro
- * @version $Id: prim.c,v 1.12 2008/12/11 10:22:51 katsu Exp $
  */
 
+#ifdef HAVE_CONFIG_H
+#include "config.h"
+#endif /* HAVE_CONFIG_H */
+
 #include <stdlib.h>
 #include <stdio.h>
 #include <string.h>
@@ -19,28 +22,18 @@
 #include <sys/time.h>
 #include <sys/types.h>
 #include <sys/stat.h>
-
-#ifdef HAVE_CONFIG_H
-#include "config.h"
-#endif /* HAVE_CONFIG_H */
+#include <fenv.h>
 
 #ifdef HAVE_CONFIG_H
 #if !HAVE_DECL_FPCLASSIFY && !HAVE_DECL_ISINF
 #include <float.h>
 #endif /* HAVE_DECL_ISINF */
 #endif /* HAVE_CONFIG_H */
-#if !defined(HAVE_CONFIG_H) || defined(HAVE_SYS_TIMES_H)
 #include <sys/times.h>
-#endif /* HAVE_SYS_TIMES_H */
-#if defined(HAVE_CONFIG_H) && !defined(HAVE_UTIMES) && defined(HAVE_UTIME_H)
+#include <sys/resource.h>
 #include <utime.h>
-#endif /* HAVE_UTIME_H */
-#if !defined(HAVE_CONFIG_H) || defined(HAVE_POLL_H)
 #include <poll.h>
-#endif /* HAVE_POLL_H */
-#if !defined(HAVE_CONFIG_H) || defined(HAVE_DLFCN_H)
 #include <dlfcn.h>
-#endif /* HAVE_DLFCN_H */
 
 #ifdef MINGW32
 #include <windows.h>
@@ -287,114 +280,20 @@ int prim_fegetround()
 }
 
 int
-prim_String_size(const char *str)
+sml_memcmp(const char *s1, int i1, const char *s2, int i2, int len)
 {
-	/* used for not only CharVector but CharArray */
-	ASSERT(OBJ_TYPE(str) == OBJTYPE_UNBOXED_VECTOR
-	       || OBJ_TYPE(str) == OBJTYPE_UNBOXED_ARRAY);
-	return OBJ_STR_SIZE(str);
-}
-
-void
-prim_String_update(char *str, int index, char ch)
-{
-	/* used for not only CharVector but CharArray */
-	ASSERT(OBJ_TYPE(str) == OBJTYPE_UNBOXED_ARRAY
-	       || OBJ_TYPE(str) == OBJTYPE_UNBOXED_VECTOR);
-	ASSERT(index >= 0 && (size_t)index < OBJ_STR_SIZE(str));
-	str[index] = ch;
-}
-
-char
-prim_String_sub(const char *str, int n)
-{
-	/* used for not only CharVector but CharArray */
-	ASSERT(OBJ_TYPE(str) == OBJTYPE_UNBOXED_ARRAY
-	       || OBJ_TYPE(str) == OBJTYPE_UNBOXED_VECTOR);
-	ASSERT(n >= 0 && (size_t)n < OBJ_STR_SIZE(str));
-	return str[n];
-}
-
-STRING
-prim_String_substring(const char *str, int beg, int len)
-{
-	ASSERT(OBJ_TYPE(str) == OBJTYPE_UNBOXED_VECTOR);
-	ASSERT(beg >= 0 && len >= 0);
-	ASSERT((size_t)(beg + len) <= OBJ_STR_SIZE(str));
-
-	return sml_str_new2(&str[beg], len);
-}
-
-int
-prim_String_cmp(const char *str1, const char *str2)
-{
-	int len1, len2, len, cmp;
-
-	ASSERT(OBJ_TYPE(str1) == OBJTYPE_UNBOXED_VECTOR);
-	ASSERT(OBJ_TYPE(str2) == OBJTYPE_UNBOXED_VECTOR);
-	len1 = OBJ_STR_SIZE(str1);
-	len2 = OBJ_STR_SIZE(str2);
-
-	len = len1 < len2 ? len1 : len2;
-	cmp = memcmp(str1, str2, len1);
-
-	if (cmp == 0) {
-		/* this is OK because both len1 and len2 are signed integer
-		 * but never negative. */
-		return len1 - len2;
+	s1 += i1;
+	s2 += i2;
+	if (s1 == s2)
+		return 0;
+	while (len > 0) {
+		unsigned char c1 = *s1;
+		unsigned char c2 = *s2;
+		if (c1 != c2)
+			return (int)c1 - (int)c2;
+		len--, s1++, s2++;
 	}
-	return cmp;
-}
-
-STRING
-prim_String_allocateMutableNoInit(unsigned int len)
-{
-	char *obj = sml_obj_alloc(OBJTYPE_UNBOXED_ARRAY, len + 1);
-	obj[len] = '\0';
-	return obj;
-}
-
-STRING
-prim_String_allocateImmutableNoInit(unsigned int len)
-{
-	char *obj = sml_obj_alloc(OBJTYPE_UNBOXED_VECTOR, len + 1);
-	obj[len] = '\0';
-	return obj;
-}
-
-STRING
-prim_String_allocateMutable(int len, char ch)
-{
-	void *obj;
-	ASSERT(len >= 0);
-	obj = prim_String_allocateMutableNoInit(len);
-	memset(obj, ch, len);
-	return obj;
-}
-
-STRING
-prim_String_allocateImmutable(int len, char ch)
-{
-	void *obj;
-	ASSERT(len >= 0);
-	obj = prim_String_allocateImmutableNoInit(len);
-	memset(obj, ch, len);
-	return obj;
-}
-
-void
-prim_String_copy(const char *src, int si, char *dst, int di, int len)
-{
-	/* used for not only CharVector but CharArray */
-	ASSERT(OBJ_TYPE(src) == OBJTYPE_UNBOXED_ARRAY
-	       || OBJ_TYPE(src) == OBJTYPE_UNBOXED_VECTOR);
-	ASSERT(OBJ_TYPE(dst) == OBJTYPE_UNBOXED_ARRAY
-	       || OBJ_TYPE(dst) == OBJTYPE_UNBOXED_VECTOR);
-	ASSERT(len >= 0);
-	ASSERT(si >= 0 && (size_t)(si + len) <= OBJ_STR_SIZE(src));
-	ASSERT(di >= 0 && (size_t)(di + len) <= OBJ_STR_SIZE(dst));
-
-	memcpy(dst + di, src + si, len);
+	return 0;
 }
 
 static unsigned int
@@ -429,28 +328,6 @@ fmt_int(int value, unsigned int radix)
 	return sml_str_new2(&buf[i], sizeof(buf) - i);
 }
 
-static STRING
-fmt_word(unsigned int value, unsigned int radix)
-{
-	char buf[sizeof(unsigned int) * CHAR_BIT + sizeof("")];
-	unsigned int i = sizeof(buf);
-
-	i = fmt(value, radix, buf, i);
-	return sml_str_new2(&buf[i], sizeof(buf) - i);
-}
-
-STRING
-prim_Int_toString(int value)
-{
-	return fmt_int(value, 10);
-}
-
-STRING
-prim_Word_toString(unsigned int value)
-{
-	return fmt_word(value, 16);
-}
-
 #define IEEEREAL_CLASS_SNAN     1   /* signaling NaN */
 #define IEEEREAL_CLASS_QNAN     2   /* quiet NaN */
 #define IEEEREAL_CLASS_INF      3   /* infinity */
@@ -544,44 +421,66 @@ prim_IntInf_toString(sml_intinf_t *n)
 {
 	char *buf, *ret;
 
-	ASSERT(OBJ_TYPE(n) == OBJTYPE_INTINF);
+	assert(OBJ_TYPE(n) == OBJTYPE_INTINF);
 	buf = sml_intinf_fmt(n, 10);
 	ret = sml_str_new(buf);
 	free(buf);
 	return ret;
 }
 
-int
-prim_IntInf_toInt(sml_intinf_t *obj)
-{
-	ASSERT(OBJ_TYPE(obj) == OBJTYPE_INTINF);
-	return sml_intinf_get_si(obj);
-}
-
 unsigned int
 prim_IntInf_toWord(sml_intinf_t *obj)
 {
-	unsigned long n;
+	assert(OBJ_TYPE(obj) == OBJTYPE_INTINF);
 
-	ASSERT(OBJ_TYPE(obj) == OBJTYPE_INTINF);
+	if (sizeof(long int) > sizeof(int)) {
+		/* if long int is larger than int, the low-order bits of
+		 * mpz_get_si is the answer. */
+		return (int)(unsigned int)(unsigned long int)
+			(sml_intinf_get_si(obj));
+	} else {
+		/* otherwise, compute 2's complement of the absolute of
+		 * the given value (mpz_get_ui) */
+		unsigned long int n = sml_intinf_get_ui(obj);
+		if (sml_intinf_sign(obj) < 0) n = -n;
+		return n;
+	}
+}
 
-	/* mpz_get_ui(op) returns least significant bits of absolute value
-	 * of "op" but this primitive requires to return least significant
-	 * bits of 2's complement form of "op". So we take 2's complement
-	 * of the return value of mpz_get_ui if "op" is negative.
-	 */
-	n = sml_intinf_get_ui(obj);
-	if (sml_intinf_sign(obj) < 0)
-		n = ~n + 1;
+uint64_t
+prim_IntInf_toWord64(sml_intinf_t *obj)
+{
+	assert(OBJ_TYPE(obj) == OBJTYPE_INTINF);
 
-	return n;
+	if (sizeof(long int) > sizeof(uint64_t)) {
+		return (uint64_t)(unsigned long int)(sml_intinf_get_si(obj));
+	} else if (sizeof(long int) == sizeof(uint64_t)) {
+		unsigned long int n = (signed long int)sml_intinf_get_ui(obj);
+		if (sml_intinf_sign(obj) < 0) n = -n;
+		return n;
+	} else {
+		uint64_t *p, n;
+		p = mpz_export(NULL, NULL, -1, sizeof(n), 0, 0, obj->value);
+		if (p == NULL) return 0;
+		n = *p;
+		free(p);
+		if (sml_intinf_sign(obj) < 0) n = -n;
+		return n;
+	}
+}
+
+double
+prim_IntInf_toReal(sml_intinf_t *obj)
+{
+	assert(OBJ_TYPE(obj) == OBJTYPE_INTINF);
+	return sml_intinf_get_d(obj);
 }
 
 sml_intinf_t *
 prim_IntInf_fromInt(int x)
 {
 	sml_intinf_t *n = sml_intinf_new();
-	sml_intinf_set_si(n, x);
+	sml_intinf_set_si(n, (long int)(unsigned long int)x);
 	return n;
 }
 
@@ -594,10 +493,38 @@ prim_IntInf_fromWord(unsigned int x)
 }
 
 sml_intinf_t *
-prim_IntInf_load(const char *src)
+prim_IntInf_fromInt64(int64_t x)
 {
 	sml_intinf_t *n = sml_intinf_new();
-	sml_intinf_set_str(n, src, 10);
+
+	if (sizeof(long int) >= sizeof(int64_t)) {
+		sml_intinf_set_si(n, (long int)(uint64_t)x);
+	} else {
+		uint64_t u = x;
+		if (x < 0) u = -u;
+		mpz_import(n->value, 1, -1, sizeof(u), 0, 0, &u);
+		if (x < 0) sml_intinf_neg(n, n);
+	}
+	return n;
+}
+
+sml_intinf_t *
+prim_IntInf_fromWord64(uint64_t x)
+{
+	sml_intinf_t *n = sml_intinf_new();
+
+	if (sizeof(long int) >= sizeof(int64_t))
+		sml_intinf_set_ui(n, (long int)(uint64_t)x);
+	else
+		mpz_import(n->value, 1, -1, sizeof(x), 0, 0, &x);
+	return n;
+}
+
+sml_intinf_t *
+prim_IntInf_fromReal(double x)
+{
+	sml_intinf_t *n = sml_intinf_new();
+	sml_intinf_set_d(n, x);
 	return n;
 }
 
@@ -605,7 +532,7 @@ sml_intinf_t *
 prim_IntInf_abs(sml_intinf_t *x)
 {
 	sml_intinf_t xv, *z;
-	ASSERT(OBJ_TYPE(x) == OBJTYPE_INTINF);
+	assert(OBJ_TYPE(x) == OBJTYPE_INTINF);
 
 	xv = *x; /* rescue from garbage collector */
 	z = sml_intinf_new();
@@ -617,8 +544,8 @@ sml_intinf_t *
 prim_IntInf_add(sml_intinf_t *x, sml_intinf_t *y)
 {
 	sml_intinf_t xv, yv, *z;
-	ASSERT(OBJ_TYPE(x) == OBJTYPE_INTINF);
-	ASSERT(OBJ_TYPE(y) == OBJTYPE_INTINF);
+	assert(OBJ_TYPE(x) == OBJTYPE_INTINF);
+	assert(OBJ_TYPE(y) == OBJTYPE_INTINF);
 
 	xv = *x, yv = *y; /* rescue from garbage collector */
 	z = sml_intinf_new();
@@ -630,8 +557,8 @@ sml_intinf_t *
 prim_IntInf_sub(sml_intinf_t *x, sml_intinf_t *y)
 {
 	sml_intinf_t xv, yv, *z;
-	ASSERT(OBJ_TYPE(x) == OBJTYPE_INTINF);
-	ASSERT(OBJ_TYPE(y) == OBJTYPE_INTINF);
+	assert(OBJ_TYPE(x) == OBJTYPE_INTINF);
+	assert(OBJ_TYPE(y) == OBJTYPE_INTINF);
 
 	xv = *x, yv = *y; /* rescue from garbage collector */
 	z = sml_intinf_new();
@@ -643,7 +570,7 @@ sml_intinf_t *
 prim_IntInf_neg(sml_intinf_t *x)
 {
 	sml_intinf_t xv, *z;
-	ASSERT(OBJ_TYPE(x) == OBJTYPE_INTINF);
+	assert(OBJ_TYPE(x) == OBJTYPE_INTINF);
 
 	xv = *x; /* rescue from garbage collector */
 	z = sml_intinf_new();
@@ -655,8 +582,8 @@ sml_intinf_t *
 prim_IntInf_mul(sml_intinf_t *x, sml_intinf_t *y)
 {
 	sml_intinf_t xv, yv, *z;
-	ASSERT(OBJ_TYPE(x) == OBJTYPE_INTINF);
-	ASSERT(OBJ_TYPE(y) == OBJTYPE_INTINF);
+	assert(OBJ_TYPE(x) == OBJTYPE_INTINF);
+	assert(OBJ_TYPE(y) == OBJTYPE_INTINF);
 
 	xv = *x, yv = *y; /* rescue from garbage collector */
 	z = sml_intinf_new();
@@ -668,8 +595,8 @@ sml_intinf_t *
 prim_IntInf_div(sml_intinf_t *x, sml_intinf_t *y)
 {
 	sml_intinf_t xv, yv, *z;
-	ASSERT(OBJ_TYPE(x) == OBJTYPE_INTINF);
-	ASSERT(OBJ_TYPE(y) == OBJTYPE_INTINF);
+	assert(OBJ_TYPE(x) == OBJTYPE_INTINF);
+	assert(OBJ_TYPE(y) == OBJTYPE_INTINF);
 
 	xv = *x, yv = *y; /* rescue from garbage collector */
 	z = sml_intinf_new();
@@ -681,8 +608,8 @@ sml_intinf_t *
 prim_IntInf_mod(sml_intinf_t *x, sml_intinf_t *y)
 {
 	sml_intinf_t xv, yv, *z;
-	ASSERT(OBJ_TYPE(x) == OBJTYPE_INTINF);
-	ASSERT(OBJ_TYPE(y) == OBJTYPE_INTINF);
+	assert(OBJ_TYPE(x) == OBJTYPE_INTINF);
+	assert(OBJ_TYPE(y) == OBJTYPE_INTINF);
 
 	xv = *x, yv = *y; /* rescue from garbage collector */
 	z = sml_intinf_new();
@@ -694,8 +621,8 @@ sml_intinf_t *
 prim_IntInf_quot(sml_intinf_t *x, sml_intinf_t *y)
 {
 	sml_intinf_t xv, yv, *z;
-	ASSERT(OBJ_TYPE(x) == OBJTYPE_INTINF);
-	ASSERT(OBJ_TYPE(y) == OBJTYPE_INTINF);
+	assert(OBJ_TYPE(x) == OBJTYPE_INTINF);
+	assert(OBJ_TYPE(y) == OBJTYPE_INTINF);
 
 	xv = *x, yv = *y; /* rescue from garbage collector */
 	z = sml_intinf_new();
@@ -707,8 +634,8 @@ sml_intinf_t *
 prim_IntInf_rem(sml_intinf_t *x, sml_intinf_t *y)
 {
 	sml_intinf_t xv, yv, *z;
-	ASSERT(OBJ_TYPE(x) == OBJTYPE_INTINF);
-	ASSERT(OBJ_TYPE(y) == OBJTYPE_INTINF);
+	assert(OBJ_TYPE(x) == OBJTYPE_INTINF);
+	assert(OBJ_TYPE(y) == OBJTYPE_INTINF);
 
 	xv = *x, yv = *y; /* rescue from garbage collector */
 	z = sml_intinf_new();
@@ -719,8 +646,8 @@ prim_IntInf_rem(sml_intinf_t *x, sml_intinf_t *y)
 int
 prim_IntInf_cmp(sml_intinf_t *x, sml_intinf_t *y)
 {
-	ASSERT(OBJ_TYPE(x) == OBJTYPE_INTINF);
-	ASSERT(OBJ_TYPE(y) == OBJTYPE_INTINF);
+	assert(OBJ_TYPE(x) == OBJTYPE_INTINF);
+	assert(OBJ_TYPE(y) == OBJTYPE_INTINF);
 	return sml_intinf_cmp(x, y);
 }
 
@@ -728,8 +655,8 @@ sml_intinf_t *
 prim_IntInf_orb(sml_intinf_t *x, sml_intinf_t *y)
 {
 	sml_intinf_t xv, yv, *z;
-	ASSERT(OBJ_TYPE(x) == OBJTYPE_INTINF);
-	ASSERT(OBJ_TYPE(y) == OBJTYPE_INTINF);
+	assert(OBJ_TYPE(x) == OBJTYPE_INTINF);
+	assert(OBJ_TYPE(y) == OBJTYPE_INTINF);
 
 	xv = *x, yv = *y; /* rescue from garbage collector */
 	z = sml_intinf_new();
@@ -741,8 +668,8 @@ sml_intinf_t *
 prim_IntInf_xorb(sml_intinf_t *x, sml_intinf_t *y)
 {
 	sml_intinf_t xv, yv, *z;
-	ASSERT(OBJ_TYPE(x) == OBJTYPE_INTINF);
-	ASSERT(OBJ_TYPE(y) == OBJTYPE_INTINF);
+	assert(OBJ_TYPE(x) == OBJTYPE_INTINF);
+	assert(OBJ_TYPE(y) == OBJTYPE_INTINF);
 
 	xv = *x, yv = *y; /* rescue from garbage collector */
 	z = sml_intinf_new();
@@ -754,8 +681,8 @@ sml_intinf_t *
 prim_IntInf_andb(sml_intinf_t *x, sml_intinf_t *y)
 {
 	sml_intinf_t xv, yv, *z;
-	ASSERT(OBJ_TYPE(x) == OBJTYPE_INTINF);
-	ASSERT(OBJ_TYPE(y) == OBJTYPE_INTINF);
+	assert(OBJ_TYPE(x) == OBJTYPE_INTINF);
+	assert(OBJ_TYPE(y) == OBJTYPE_INTINF);
 
 	xv = *x, yv = *y; /* rescue from garbage collector */
 	z = sml_intinf_new();
@@ -767,7 +694,7 @@ sml_intinf_t *
 prim_IntInf_notb(sml_intinf_t *x)
 {
 	sml_intinf_t xv, *z;
-	ASSERT(OBJ_TYPE(x) == OBJTYPE_INTINF);
+	assert(OBJ_TYPE(x) == OBJTYPE_INTINF);
 
 	xv = *x; /* rescue from garbage collector */
 	z = sml_intinf_new();
@@ -779,8 +706,8 @@ sml_intinf_t *
 prim_IntInf_pow(sml_intinf_t *x, int e)
 {
 	sml_intinf_t xv, *z;
-	ASSERT(OBJ_TYPE(x) == OBJTYPE_INTINF);
-	ASSERT(e >= 0);
+	assert(OBJ_TYPE(x) == OBJTYPE_INTINF);
+	assert(e >= 0);
 
 	xv = *x; /* rescue from garbage collector */
 	z = sml_intinf_new();
@@ -792,7 +719,7 @@ int
 prim_IntInf_log2(sml_intinf_t *x)
 {
 	sml_intinf_t xv;
-	ASSERT(OBJ_TYPE(x) == OBJTYPE_INTINF);
+	assert(OBJ_TYPE(x) == OBJTYPE_INTINF);
 
 	xv = *x; /* rescue from garbage collector */
 	return sml_intinf_log2(&xv);
@@ -804,8 +731,8 @@ prim_Time_gettimeofday(int *ret)
 	struct timeval tv;
 	int err;
 
-	ASSERT(OBJ_TYPE(ret) == OBJTYPE_UNBOXED_ARRAY);
-	ASSERT(OBJ_SIZE(ret) >= sizeof(int) * 2);
+	assert(OBJ_TYPE(ret) == OBJTYPE_UNBOXED_ARRAY);
+	assert(OBJ_SIZE(ret) >= sizeof(int) * 2);
 
 	err = gettimeofday(&tv, NULL);
 	ret[0] = tv.tv_sec;
@@ -816,13 +743,28 @@ prim_Time_gettimeofday(int *ret)
 int
 prim_Timer_getTimes(int *ret)
 {
-#ifdef HAVE_TIMES
+#if defined HAVE_GETRUSAGE
+	struct rusage r;
+	int err;
+
+	err = getrusage(RUSAGE_SELF, &r);
+	if (err == 0) {
+		ret[0] = r.ru_stime.tv_sec;
+		ret[1] = r.ru_stime.tv_usec;
+		ret[2] = r.ru_utime.tv_sec;
+		ret[3] = r.ru_utime.tv_usec;
+		/* FIXME: do we put GC time still here? */
+		ret[4] = 0;  /* GC seconds */
+		ret[5] = 0;  /* GC microseconds */
+	}
+	return err;
+#elif defined HAVE_TIMES
 	struct tms tms;
 	static long clocks_per_sec = 0;
 	clock_t clk;
 
-	ASSERT(OBJ_TYPE(ret) == OBJTYPE_UNBOXED_ARRAY);
-	ASSERT(OBJ_SIZE(ret) >= sizeof(int) * 6);
+	assert(OBJ_TYPE(ret) == OBJTYPE_UNBOXED_ARRAY);
+	assert(OBJ_SIZE(ret) >= sizeof(int) * 6);
 
 	if (clocks_per_sec == 0)
 		clocks_per_sec = sysconf(_SC_CLK_TCK);
@@ -841,8 +783,8 @@ prim_Timer_getTimes(int *ret)
 	struct timeval tv;
 	int err;
 
-	ASSERT(OBJ_TYPE(ret) == OBJTYPE_UNBOXED_ARRAY);
-	ASSERT(OBJ_SIZE(ret) >= sizeof(int) * 6);
+	assert(OBJ_TYPE(ret) == OBJTYPE_UNBOXED_ARRAY);
+	assert(OBJ_SIZE(ret) >= sizeof(int) * 6);
 
 	err = gettimeofday(&tv, NULL);
 	ret[0] = 0;  /* sys seconds */
@@ -858,53 +800,24 @@ prim_Timer_getTimes(int *ret)
 
 unsigned int
 prim_Date_strfTime(char *buf, unsigned int maxsize, const char *format,
-		   int sec, int min, int hour, int mday, int month,
-		   int year, int wday, int yday, int isdst)
+		   const struct tm *tm)
 {
-	struct tm tm;
-	tm.tm_sec = sec;
-	tm.tm_min = min;
-	tm.tm_hour = hour;
-	tm.tm_mday = mday;
-	tm.tm_mon = month;
-	tm.tm_year = year;
-	tm.tm_wday = wday;
-	tm.tm_yday = yday;
-	tm.tm_isdst = isdst;
-	return strftime(buf, maxsize, format, &tm);
-}
-
-char *
-prim_Date_ascTime(int sec, int min, int hour, int mday, int month, int year,
-		  int wday, int yday, int isdst)
-{
-	struct tm tm;
-	tm.tm_sec = sec;
-	tm.tm_min = min;
-	tm.tm_hour = hour;
-	tm.tm_mday = mday;
-	tm.tm_mon = month;
-	tm.tm_year = year;
-	tm.tm_wday = wday;
-	tm.tm_yday = yday;
-	tm.tm_isdst = isdst;
-	return asctime(&tm);
+	return strftime(buf, maxsize, format, tm);
 }
 
 int
-prim_Date_mkTime(int sec, int min, int hour, int mday, int month, int year,
-                 int wday, int yday, int isdst)
+prim_Date_mkTime(const struct tm *time)
 {
-	struct tm tm;
-	tm.tm_sec = sec;
-	tm.tm_min = min;
-	tm.tm_hour = hour;
-	tm.tm_mday = mday;
-	tm.tm_mon = month;
-	tm.tm_year = year;
-	tm.tm_wday = wday;
-	tm.tm_yday = yday;
-	tm.tm_isdst = isdst;
+        struct tm tm;
+	tm.tm_sec = time->tm_sec;
+	tm.tm_min = time->tm_min;
+	tm.tm_hour = time->tm_hour;
+	tm.tm_mday = time->tm_mday;
+	tm.tm_mon = time->tm_mon;
+	tm.tm_year = time->tm_year;
+	tm.tm_wday = time->tm_wday;
+	tm.tm_yday = time->tm_yday;
+	tm.tm_isdst = time->tm_isdst;
 	return mktime(&tm);
 }
 
@@ -946,139 +859,39 @@ prim_Date_gmTime(int time, int *ret)
 	return 0;
 }
 
-double
-prim_Pack_packReal64Little(unsigned char byte0, unsigned char byte1,
-			   unsigned char byte2, unsigned char byte3,
-			   unsigned char byte4, unsigned char byte5,
-			   unsigned char byte6, unsigned char byte7)
-{
-	double result;
-
-#ifdef WORDS_BIGENDIAN
-	char src[8] = {byte7, byte6, byte5, byte4, byte3, byte2, byte1, byte0};
-#else
-	char src[8] = {byte0, byte1, byte2, byte3, byte4, byte5, byte6, byte7};
-#endif /* WORDS_BIGENDIAN */
-
-	memcpy(&result, src, sizeof(double) < 8 ? sizeof(double) : 8);
-	return result;
-}
-
-double
-prim_Pack_packReal64Big(unsigned char byte0, unsigned char byte1,
-			unsigned char byte2, unsigned char byte3,
-			unsigned char byte4, unsigned char byte5,
-			unsigned char byte6, unsigned char byte7)
-{
-	double result;
-
-#ifdef WORDS_BIGENDIAN
-	char src[8] = {byte0, byte1, byte2, byte3, byte4, byte5, byte6, byte7};
-#else
-	char src[8] = {byte7, byte6, byte5, byte4, byte3, byte2, byte1, byte0};
-#endif /* WORDS_BIGENDIAN */
-
-	memcpy(&result, src, sizeof(double) < 8 ? sizeof(double) : 8);
-	return result;
-}
-
-void
-prim_Pack_unpackReal64Little(double d, unsigned char *buf)
-{
-#ifdef WORDS_BIGENDIAN
-	size_t i, len = sizeof(double) < 8 ? sizeof(double) : 8;
-	for (i = 0; i < len; i++)
-		buf[i] = ((unsigned char *)&d)[len - i - 1];
-#else
-	memcpy(buf, &d, sizeof(double) < 8 ? sizeof(double) : 8);
-#endif /* WORDS_BIGENDIAN */
-}
-
-void
-prim_Pack_packReal32Little(unsigned char byte0, unsigned char byte1,
-			   unsigned char byte2, unsigned char byte3,
-			   float *ret)
-{
-#ifdef WORDS_BIGENDIAN
-	char src[4] = {byte3, byte2, byte1, byte0};
-#else
-	char src[4] = {byte0, byte1, byte2, byte3};
-#endif /* WORDS_BIGENDIAN */
-
-	memcpy(ret, src, sizeof(float) < 4 ? sizeof(float) : 4);
-}
-
-void
-prim_Pack_packReal32Big(unsigned char byte0, unsigned char byte1,
-			unsigned char byte2, unsigned char byte3,
-                        float *ret)
-{
-#ifdef WORDS_BIGENDIAN
-	char src[4] = {byte0, byte1, byte2, byte3};
-#else
-	char src[4] = {byte3, byte2, byte1, byte0};
-#endif /* WORDS_BIGENDIAN */
-
-	memcpy(ret, src, sizeof(float) < 4 ? sizeof(float) : 4);
-}
-
-void
-prim_Pack_unpackReal32Little(float d, unsigned char *buf)
-{
-#ifdef WORDS_BIGENDIAN
-	size_t i, len = sizeof(float) < 4 ? sizeof(float) : 4;
-	for (i = 0; i < len; i++)
-		buf[i] = ((unsigned char *)&d)[len - i - 1];
-#else
-	memcpy(buf, &d, sizeof(float) < 4 ? sizeof(float) : 4);
-#endif /* WORDS_BIGENDIAN */
-}
-
-
-/* HERE */
-
-
 int
 prim_StandardC_errno()
 {
 	return errno;
 }
 
-int
-prim_cconst_int(const char *name)
-{
+#define PRIM_CONST_FUNC(ty, const) \
+	ty prim_const_##const() { return const; }
+#define PRIM_CONST_FUNC_DUMMY(ty, const) \
+	ty prim_const_##const() { return 0; }
+
 #ifdef HAVE_DLOPEN
-	if (strcmp(name, "RTLD_LAZY") == 0)
-		return RTLD_LAZY;
-	if (strcmp(name, "RTLD_NOW") == 0)
-		return RTLD_NOW;
-	if (strcmp(name, "RTLD_GLOBAL") == 0)
-		return RTLD_GLOBAL;
-	if (strcmp(name, "RTLD_LOCAL") == 0)
-		return RTLD_LOCAL;
+PRIM_CONST_FUNC(int, RTLD_LAZY)
+PRIM_CONST_FUNC(int, RTLD_NOW)
+PRIM_CONST_FUNC(int, RTLD_LOCAL)
+PRIM_CONST_FUNC(int, RTLD_GLOBAL)
+PRIM_CONST_FUNC(void *, RTLD_DEFAULT)
+PRIM_CONST_FUNC(void *, RTLD_NEXT)
+#else
+PRIM_CONST_FUNC_DUMMY(int, RTLD_LAZY)
+PRIM_CONST_FUNC_DUMMY(int, RTLD_NOW)
+PRIM_CONST_FUNC_DUMMY(int, RTLD_LOCAL)
+PRIM_CONST_FUNC_DUMMY(int, RTLD_GLOBAL)
+PRIM_CONST_FUNC_DUMMY(void *, RTLD_DEFAULT)
+PRIM_CONST_FUNC_DUMMY(void *, RTLD_NEXT)
 #endif /* HAVE_DLOPEN */
-	if (strcmp(name, "SEEK_SET") == 0)
-		return SEEK_SET;
-	if (strcmp(name, "SEEK_CUR") == 0)
-		return SEEK_CUR;
-#ifdef FE_TONEAREST
-	if (strcmp(name, "FE_TONEAREST") == 0)
-		return FE_TONEAREST;
-#endif /* FE_TONEAREST */
-#ifdef FE_DOWNWARD
-	if (strcmp(name, "FE_DOWNWARD") == 0)
-		return FE_DOWNWARD;
-#endif /* FE_DOWNWARD */
-#ifdef FE_UPWARD
-	if (strcmp(name, "FE_UPWARD") == 0)
-		return FE_UPWARD;
-#endif /* FE_UPWARD */
-#ifdef FE_TOWARDZERO
-	if (strcmp(name, "FE_TOWARDZERO") == 0)
-		return FE_TOWARDZERO;
-#endif /* FE_TOWARDZERO */
-	return 0;
-}
+
+PRIM_CONST_FUNC(int, SEEK_SET)
+PRIM_CONST_FUNC(int, SEEK_CUR)
+PRIM_CONST_FUNC(int, FE_TONEAREST)
+PRIM_CONST_FUNC(int, FE_DOWNWARD)
+PRIM_CONST_FUNC(int, FE_UPWARD)
+PRIM_CONST_FUNC(int, FE_TOWARDZERO)
 
 static struct {
 	int errnum;
@@ -1240,7 +1053,7 @@ prim_GenericOS_syserror(const char *errorname)
 	unsigned int i;
 	int errnum = -1;
 
-	ASSERT(OBJ_TYPE(errorname) == OBJTYPE_UNBOXED_VECTOR);
+	assert(OBJ_TYPE(errorname) == OBJTYPE_UNBOXED_VECTOR);
 
 	/* errorname[0] always exists due to existence of sentinel. */
 	if (isdigit(errorname[0]))
@@ -1258,72 +1071,17 @@ prim_GenericOS_syserror(const char *errorname)
 void
 prim_GenericOS_exit(int status)
 {
-	/* FIXME: finalization is needed? */
-#ifdef HAVE_INTERACTIVE_MODE
-	if (interactive_mode)
-		return interact_prim_exit(status);
-#endif /* HAVE_INTERACTIVE_MODE */
-
-	exit(status);
+	sml_exit(status);
 }
 
-static void
-puts_posix(int fd, const char *buf, size_t len)
-{
-	ssize_t n;
-
-	while (len > 0) {
-		n = write(fd, buf, len);
-		if (n < 0) {
-			if (errno == EINTR)
-				continue;
-			/* give up. */
-			return;
-		}
-		buf += n;
-		len -= n;
-	}
-}
-
-void
-prim_print(const char *str)
-{
-	ASSERT(OBJ_TYPE(str) == OBJTYPE_UNBOXED_VECTOR);
-#ifdef HAVE_INTERACTIVE_MODE
-	if (interactive_mode) {
-		interact_prim_print(str);
-		return;
-	}
-#endif /* HAVE_INTERACTIVE_MODE */
-
-	puts_posix(1, str, OBJ_STR_SIZE(str));
-}
-
-#if 0
-void
-prim_printerr(const char *str)
-{
-	ASSERT(OBJ_TYPE(str) == OBJTYPE_UNBOXED_VECTOR);
-
-#ifdef HAVE_INTERACTIVE_MODE
-	if (interactive_mode) {
-		interact_prim_printerr(str);
-		return;
-	}
-#endif /* HAVE_INTERACTIVE_MODE */
-
-	puts_posix(2, str, OBJ_STR_SIZE(str));
-}
-#endif
-
 int
 prim_GenericOS_open(const char *filename, const char *fmode)
 {
 	const char *str;
 	int flags, subflags;
 
-	ASSERT(OBJ_TYPE(filename) == OBJTYPE_UNBOXED_VECTOR);
-	ASSERT(OBJ_TYPE(fmode) == OBJTYPE_UNBOXED_VECTOR);
+	assert(OBJ_TYPE(filename) == OBJTYPE_UNBOXED_VECTOR);
+	assert(OBJ_TYPE(fmode) == OBJTYPE_UNBOXED_VECTOR);
 
 	str = fmode;
 	switch (*(str++)) {
@@ -1357,14 +1115,15 @@ prim_GenericOS_open(const char *filename, const char *fmode)
 		subflags |= O_BINARY;
 #endif
 
-	return open(filename, flags | subflags, 0777);
+	return open(filename, flags | subflags, 0666);
 }
 
 int
 prim_GenericOS_read(int fd, char *buf, unsigned int offset, unsigned int len)
 {
-	ASSERT(OBJ_TYPE(buf) == OBJTYPE_UNBOXED_ARRAY);
-	ASSERT(offset + len <= OBJ_SIZE(buf));
+	assert(OBJ_TYPE(buf) == OBJTYPE_UNBOXED_ARRAY
+	       || OBJ_TYPE(buf) == OBJTYPE_UNBOXED_VECTOR);
+	assert(offset + len <= OBJ_SIZE(buf));
 
 #ifdef HAVE_INTERACTIVE_MODE
 	if (interactive_mode && fd == 0)
@@ -1378,9 +1137,9 @@ int
 prim_GenericOS_write(int fd, const char *buf,
 		     unsigned int offset, unsigned int len)
 {
-	ASSERT(OBJ_TYPE(buf) == OBJTYPE_UNBOXED_ARRAY
+	assert(OBJ_TYPE(buf) == OBJTYPE_UNBOXED_ARRAY
 	       || OBJ_TYPE(buf) == OBJTYPE_UNBOXED_VECTOR);
-	ASSERT(offset + len <= OBJ_SIZE(buf));
+	assert(offset + len <= OBJ_SIZE(buf));
 
 #ifdef HAVE_INTERACTIVE_MODE
 	if (interactive_mode && fd == 0)
@@ -1396,83 +1155,33 @@ prim_GenericOS_lseek(int fd, /*off_t*/ int offset, int whence)
 	return lseek(fd, offset, whence);
 }
 
-
-#define ML_S_IFIFO  0x1000
-#define ML_S_IFCHR  0x2000
-#define ML_S_IFDIR  0x4000
-#define ML_S_IFBLK  0x6000
-#define ML_S_IFREG  0x8000
-#define ML_S_IFLNK  0xa000
-#define ML_S_IFSOCK 0xc000
-#define ML_S_ISUID  0x0800
-#define ML_S_ISGID  0x0400
-#define ML_S_ISVTX  0x0200
-#define ML_S_IRUSR  0x0100
-#define ML_S_IWUSR  0x0080
-#define ML_S_IXUSR  0x0040
-#define ML_S_IFMT   0xf000
-
-#ifndef S_IFLNK
-#define S_IFLNK  0
-#endif /* S_IFLNK */
-#ifndef S_IFSOCK
-#define S_IFSOCK  0
-#endif /* S_IFSOCK */
-#ifndef S_ISUID
-#define S_ISUID  0
-#endif /* S_ISUID */
-#ifndef S_ISGID
-#define S_ISGID  0
-#endif /* S_ISGID */
-#ifndef S_ISVTX
-#define S_ISVTX  0
-#endif /* S_ISVTX */
+PRIM_CONST_FUNC(unsigned int, S_IFMT)
+PRIM_CONST_FUNC(unsigned int, S_IFIFO)
+PRIM_CONST_FUNC(unsigned int, S_IFCHR)
+PRIM_CONST_FUNC(unsigned int, S_IFDIR)
+PRIM_CONST_FUNC(unsigned int, S_IFBLK)
+PRIM_CONST_FUNC(unsigned int, S_IFREG)
+PRIM_CONST_FUNC(unsigned int, S_IFLNK)
+PRIM_CONST_FUNC(unsigned int, S_IFSOCK)
+PRIM_CONST_FUNC(unsigned int, S_ISUID)
+PRIM_CONST_FUNC(unsigned int, S_ISGID)
+PRIM_CONST_FUNC(unsigned int, S_ISVTX)
+PRIM_CONST_FUNC(unsigned int, S_IRUSR)
+PRIM_CONST_FUNC(unsigned int, S_IWUSR)
+PRIM_CONST_FUNC(unsigned int, S_IXUSR)
 
 static void
 set_stat(struct stat *st, unsigned int *ret)
 {
-	ASSERT(OBJ_TYPE(ret) == OBJTYPE_UNBOXED_ARRAY);
-	ASSERT(OBJ_SIZE(ret) >= sizeof(unsigned int) * 6);
+	assert(OBJ_TYPE(ret) == OBJTYPE_UNBOXED_ARRAY);
+	assert(OBJ_SIZE(ret) >= sizeof(unsigned int) * 6);
 
 	ret[0] = st->st_dev;
 	ret[1] = st->st_ino;
+	ret[2] = st->st_mode;
 	ret[3] = st->st_atime;
 	ret[4] = st->st_mtime;
 	ret[5] = st->st_size;
-
-#if S_IFIFO == ML_S_IFIFO \
-	&& S_IFCHR == ML_S_IFCHR \
-	&& S_IFDIR == ML_S_IFDIR \
-	&& S_IFBLK == ML_S_IFBLK \
-	&& S_IFREG == ML_S_IFREG \
-	&& S_IFLNK == ML_S_IFLNK \
-	&& S_IFSOCK == ML_S_IFSOCK \
-	&& S_ISUID == ML_S_ISUID \
-	&& S_ISGID == ML_S_ISGID \
-	&& S_ISVTX == ML_S_ISVTX \
-	&& S_IRUSR == ML_S_IRUSR \
-	&& S_IWUSR == ML_S_IWUSR \
-	&& S_IXUSR == ML_S_IXUSR
-	ret[2] = st->st_mode;
-#else
-	{
-		unsigned int mode = 0;
-		mode |= (st->st_mode & S_IFIFO) ? ML_S_IFIFO : 0;
-		mode |= (st->st_mode & S_IFCHR) ? ML_S_IFCHR : 0;
-		mode |= (st->st_mode & S_IFDIR) ? ML_S_IFDIR : 0;
-		mode |= (st->st_mode & S_IFBLK) ? ML_S_IFBLK : 0;
-		mode |= (st->st_mode & S_IFREG) ? ML_S_IFREG : 0;
-		mode |= (st->st_mode & S_IFLNK) ? ML_S_IFLNK : 0;
-		mode |= (st->st_mode & S_IFSOCK) ? ML_S_IFSOCK : 0;
-		mode |= (st->st_mode & S_ISUID) ? ML_S_ISUID : 0;
-		mode |= (st->st_mode & S_ISGID) ? ML_S_ISGID : 0;
-		mode |= (st->st_mode & S_ISVTX) ? ML_S_ISVTX : 0;
-		mode |= (st->st_mode & S_IRUSR) ? ML_S_IRUSR : 0;
-		mode |= (st->st_mode & S_IWUSR) ? ML_S_IWUSR : 0;
-		mode |= (st->st_mode & S_IXUSR) ? ML_S_IXUSR : 0;
-		ret[2] = mode;
-	}
-#endif
 }
 
 int
@@ -1499,6 +1208,18 @@ prim_GenericOS_stat(const char *filename, unsigned int *ret)
 	return err;
 }
 
+int
+prim_GenericOS_lstat(const char *filename, unsigned int *ret)
+{
+	int err;
+	struct stat st;
+
+	err = lstat(filename, &st);
+	if (err == 0)
+		set_stat(&st, ret);
+	return err;
+}
+
 int
 prim_GenericOS_utime(const char *filename, unsigned int atime,
 		     unsigned int mtime)
@@ -1507,7 +1228,7 @@ prim_GenericOS_utime(const char *filename, unsigned int atime,
 	struct timeval times[2];
 
 	/* FIXME: untested */
-	ASSERT(OBJ_TYPE(filename) == OBJTYPE_UNBOXED_VECTOR);
+	assert(OBJ_TYPE(filename) == OBJTYPE_UNBOXED_VECTOR);
 	times[0].tv_sec = atime;
 	times[0].tv_usec = 0;
 	times[1].tv_sec = mtime;
@@ -1532,7 +1253,7 @@ prim_GenericOS_readlink(const char *filename)
 	ssize_t n, len;
 	void *obj;
 
-	ASSERT(OBJ_TYPE(filename) == OBJTYPE_UNBOXED_VECTOR);
+	assert(OBJ_TYPE(filename) == OBJTYPE_UNBOXED_VECTOR);
 
 	n = readlink(filename, buf, sizeof(buf));
 	if (n < 0)
@@ -1562,7 +1283,7 @@ prim_GenericOS_readlink(const char *filename)
 int
 prim_GenericOS_chdir(const char *dirname)
 {
-	ASSERT(OBJ_TYPE(dirname) == OBJTYPE_UNBOXED_VECTOR);
+	assert(OBJ_TYPE(dirname) == OBJTYPE_UNBOXED_VECTOR);
 
 #ifdef HAVE_INTERACTIVE_MODE
 	if (interactive_mode)
@@ -1575,7 +1296,7 @@ prim_GenericOS_chdir(const char *dirname)
 int
 prim_GenericOS_mkdir(const char *dirname, /*mode_t*/ int mode)
 {
-	ASSERT(OBJ_TYPE(dirname) == OBJTYPE_UNBOXED_VECTOR);
+	assert(OBJ_TYPE(dirname) == OBJTYPE_UNBOXED_VECTOR);
 #ifdef MINGW32
 	return _mkdir(dirname);
 #else
@@ -1603,7 +1324,7 @@ prim_GenericOS_getcwd()
 /*DIR**/ void *
 prim_GenericOS_opendir(const char *dirname)
 {
-	ASSERT(OBJ_TYPE(dirname) == OBJTYPE_UNBOXED_VECTOR);
+	assert(OBJ_TYPE(dirname) == OBJTYPE_UNBOXED_VECTOR);
 	return opendir(dirname);
 }
 
@@ -1645,9 +1366,9 @@ prim_GenericOS_poll(int *fdary, unsigned int *evary, int timeout_sec,
 	int nfds, err;
 
 	/* FIXME: untested */
-	ASSERT(OBJ_TYPE(fdary) == OBJTYPE_UNBOXED_ARRAY);
-	ASSERT(OBJ_TYPE(evary) == OBJTYPE_UNBOXED_ARRAY);
-	ASSERT(OBJ_SIZE(fdary) == OBJ_SIZE(evary));
+	assert(OBJ_TYPE(fdary) == OBJTYPE_UNBOXED_ARRAY);
+	assert(OBJ_TYPE(evary) == OBJTYPE_UNBOXED_ARRAY);
+	assert(OBJ_SIZE(fdary) == OBJ_SIZE(evary));
 
 	FD_ZERO(&infds);
 	FD_ZERO(&outfds);
@@ -1702,9 +1423,9 @@ prim_GenericOS_poll(int *fdary, unsigned int *evary, int timeout_sec,
 	int err;
 
 	/* FIXME: untested */
-	ASSERT(OBJ_TYPE(fdary) == OBJTYPE_UNBOXED_ARRAY);
-	ASSERT(OBJ_TYPE(evary) == OBJTYPE_UNBOXED_ARRAY);
-	ASSERT(OBJ_SIZE(fdary) == OBJ_SIZE(evary));
+	assert(OBJ_TYPE(fdary) == OBJTYPE_UNBOXED_ARRAY);
+	assert(OBJ_TYPE(evary) == OBJTYPE_UNBOXED_ARRAY);
+	assert(OBJ_SIZE(fdary) == OBJ_SIZE(evary));
 
 	nfds = OBJ_SIZE(fdary) / sizeof(int);
 	fds = xmalloc(nfds * sizeof(struct pollfd));
@@ -1750,149 +1471,14 @@ prim_GenericOS_poll(int *fdary, unsigned int *evary, int timeout_sec,
 #endif /* HAVE_SELECT | HAVE_POLL */
 }
 
-int
-prim_Platform_isBigEndian()
-{
-#ifdef WORDS_BIGENDIAN
-	return 1;
-#else
-	return 0;
-#endif /* WORDS_BIGENDIAN */
-}
-
-STRING
-prim_Platform_getPlatform()
-{
-#ifdef SMLSHARP_PLATFORM
-	return sml_str_new(SMLSHARP_PLATFORM);
-#else
-	return sml_str_new("");  /* dummy */
-#endif /* SMLSHARP_PLATFORM */
-}
-
-void
-prim_CopyMemory(void *dst, unsigned int doff,
-		const void *src, unsigned int soff,
-		unsigned int len, unsigned int tag)
-{
-	void **writeaddr, **srcaddr;
-	unsigned int i;
-
-	ASSERT((tag == TAG_UNBOXED
-		&& (OBJ_TYPE(dst) == OBJTYPE_UNBOXED_ARRAY
-		    || OBJ_TYPE(dst) == OBJTYPE_UNBOXED_VECTOR))
-	       || (tag == TAG_BOXED
-		   && (OBJ_TYPE(dst) == OBJTYPE_BOXED_ARRAY
-		       || OBJ_TYPE(dst) == OBJTYPE_BOXED_VECTOR)));
-	ASSERT((tag == TAG_UNBOXED
-		&& (OBJ_TYPE(src) == OBJTYPE_UNBOXED_ARRAY
-		    || OBJ_TYPE(src) == OBJTYPE_UNBOXED_VECTOR))
-	       || (tag == TAG_BOXED
-		   && (OBJ_TYPE(src) == OBJTYPE_BOXED_ARRAY
-		       || OBJ_TYPE(src) == OBJTYPE_BOXED_VECTOR)));
-	ASSERT(doff + len <= OBJ_SIZE(dst));
-	ASSERT(soff + len <= OBJ_SIZE(src));
-
-	if (tag == TAG_UNBOXED) {
-		memmove((char*)dst + doff, (char*)src + soff, len);
-	} else if (src != dst || doff < soff) {
-		writeaddr = (void**)((char*)dst + doff);
-		srcaddr = (void**)((char*)src + soff);
-		for (i = 0; i < len / sizeof(void*); i++)
-			sml_write(dst, writeaddr++, *(srcaddr++));
-	} else {
-		writeaddr = (void**)((char*)dst + doff + len);
-		srcaddr = (void**)((char*)src + soff + len);
-		for (i = 0; i < len / sizeof(void*); i++)
-			sml_write(dst, --writeaddr, *(--srcaddr));
-	}
-}
-
-int
-prim_UnmanagedMemory_subInt(void *p)
-{
-	return *(int*)p;
-}
-
-double
-prim_UnmanagedMemory_subReal(void *p)
-{
-	return *(double*)p;
-}
-
-unsigned int
-prim_UnmanagedMemory_subWord(void *p)
-{
-	return *(unsigned int*)p;
-}
-
-unsigned char
-prim_UnmanagedMemory_subByte(void *p)
-{
-	return *(unsigned char*)p;
-}
-
-void *
-prim_UnmanagedMemory_subPtr(void *p)
-{
-	return *(void**)p;
-}
-
 STRING
 prim_UnmanagedMemory_import(void *ptr, unsigned int len)
 {
-	return sml_str_new2(ptr, len);
-}
+	void *obj;
 
-void *
-prim_UnmanagedMemory_export(const char *str, unsigned int offset,
-			    unsigned int size)
-{
-	void *p;
-
-	ASSERT(OBJ_TYPE(str) == OBJTYPE_UNBOXED_VECTOR
-	       || OBJ_TYPE(str) == OBJTYPE_UNBOXED_ARRAY);
-	ASSERT(offset < OBJ_STR_SIZE(str) && size < OBJ_STR_SIZE(str) - offset);
-
-	p = xmalloc(size);
-	memcpy(p, str + offset, size);
-	return p;
-}
-
-int
-prim_UnmanagedString_size(void *ptr)
-{
-	return strlen(ptr);
-}
-
-void
-prim_UnmanagedMemory_updateByte(void *address, unsigned char value)
-{
-	*(unsigned char*)address = value;
-}
-
-void
-prim_UnmanagedMemory_updateWord(void *address, unsigned int value)
-{
-	*(unsigned int *)address = value;
-}
-
-void
-prim_UnmanagedMemory_updateInt(void *address, int value)
-{
-	*(int *)address = value;
-}
-
-void
-prim_UnmanagedMemory_updateReal(void *address, double value)
-{
-	*(double *)address = value;
-}
-
-void
-prim_UnmanagedMemory_updatePtr(void *address, void *value)
-{
-	*(void **)address = value;
+	obj = sml_obj_alloc(OBJTYPE_UNBOXED_VECTOR, len);
+	memcpy(obj, ptr, len);
+	return obj;
 }
 
 int
@@ -1909,12 +1495,6 @@ prim_CommandLine_argv(int index)
 	return sml_argv;
 }
 
-void *
-prim_xmalloc(/*size_t*/ int size)
-{
-	return xmalloc(size);
-}
-
 STRING
 prim_executable_path()
 {
